import logging
import secrets
import uuid

import constance
from django.contrib.auth import get_user_model
from django.db import models, transaction
from django.db.models import (Q, QuerySet)

from core.models.submission import MetaSubmission
from core.models.assignment import TutorSubmissionAssignment

log = logging.getLogger(__name__)
config = constance.config


class SubscriptionEnded(Exception):
    pass


class SubscriptionTemporarilyEnded(Exception):
    pass


class NotMoreThanTwoOpenAssignmentsAllowed(Exception):
    pass


def get_random_element_from_queryset(queryset):
    qs_elements = queryset.all()
    length = len(qs_elements)
    index = secrets.choice(range(length))
    return qs_elements[index]


class SubmissionSubscription(models.Model):

    RANDOM = 'random'
    STUDENT_QUERY = 'student'
    EXAM_TYPE_QUERY = 'exam'
    SUBMISSION_TYPE_QUERY = 'submission_type'

    type_query_mapper = {
        RANDOM: '__any',
        STUDENT_QUERY: 'student__pk',
        EXAM_TYPE_QUERY: 'student__exam__pk',
        SUBMISSION_TYPE_QUERY: 'type__pk',
    }

    QUERY_CHOICE = (
        (RANDOM, 'Query for any submission'),
        (STUDENT_QUERY, 'Query for submissions of student'),
        (EXAM_TYPE_QUERY, 'Query for submissions of exam type'),
        (SUBMISSION_TYPE_QUERY, 'Query for submissions of submissions_type'),
    )

    FEEDBACK_CREATION = 'feedback-creation'
    FEEDBACK_VALIDATION = 'feedback-validation'
    FEEDBACK_CONFLICT_RESOLUTION = 'feedback-conflict-resolution'

    assignment_count_on_stage = {
        FEEDBACK_CREATION: 0,
        FEEDBACK_VALIDATION: 1,
        FEEDBACK_CONFLICT_RESOLUTION: 2,
    }

    stages = (
        (FEEDBACK_CREATION, 'No feedback was ever assigned'),
        (FEEDBACK_VALIDATION, 'Feedback exists but is not validated'),
        (FEEDBACK_CONFLICT_RESOLUTION, 'Previous correctors disagree'),
    )

    deactivated = models.BooleanField(default=False)
    subscription_id = models.UUIDField(primary_key=True,
                                       default=uuid.uuid4,
                                       editable=False)
    owner = models.ForeignKey(get_user_model(),
                              on_delete=models.CASCADE,
                              related_name='subscriptions')
    query_key = models.UUIDField(null=True)
    query_type = models.CharField(max_length=75,
                                  choices=QUERY_CHOICE,
                                  default=RANDOM)
    feedback_stage = models.CharField(choices=stages,
                                      max_length=40,
                                      default=FEEDBACK_CREATION)

    class Meta:
        unique_together = ('owner',
                           'query_key',
                           'query_type',
                           'feedback_stage')

    def _get_submission_base_query(self) -> QuerySet:
        """ Get all submissions that are filtered by the query key and type,
        e.g. all submissions of one student or submission type.
        """
        if self.query_type == self.RANDOM:
            return MetaSubmission.objects.all()

        return MetaSubmission.objects.filter(
            **{'submission__' + self.type_query_mapper[self.query_type]:
               self.query_key})

    def _get_submissions_that_do_not_have_final_feedback(self) -> QuerySet:
        """ There are a number of conditions to check for each submission

        1. The submission does not have final feedback
        2. The submission was not shown to this user before
        3. The submission is not currently assigned to somebody else

        Returns:
            QuerySet -- a list of all submissions ready for consumption
        """
        return self._get_submission_base_query() \
            .select_for_update(of=('self',)).exclude(
            Q(has_final_feedback=True) |
            Q(has_active_assignment=True) |
            Q(feedback_authors=self.owner)
        )

    def _get_available_submissions_in_subscription_stage(self) -> QuerySet:
        """ Another filter this time it returns all the submissions that
        are valid in this stage. That means all previous stages have been
        completed.

        Raises:
            SubscriptionEnded -- if the subscription will not yield
                                 subscriptions in the future
            SubscriptionTemporarilyEnded -- wait until new become available
        """
        candidates = self._get_submissions_that_do_not_have_final_feedback()

        if candidates.count() == 0:
            raise SubscriptionEnded(
                f'The task which user {self.owner} subscribed to is done')

        done_assignments_count = self.assignment_count_on_stage[self.feedback_stage]  # noqa
        stage_candidates = candidates.filter(
            done_assignments=done_assignments_count,
        )

        if stage_candidates.count() == 0:
            raise SubscriptionTemporarilyEnded(
                'Currently unavailable. Please check for more soon. '
                'Submissions remaining: %s' % stage_candidates.count())

        if (config.STOP_ON_PASS and
                self.feedback_stage == self.FEEDBACK_CREATION):
            stage_candidates = stage_candidates.exclude(
                Q(submission__student__passes_exam=True) &
                Q(submission__student__exam__pass_only=True)
            )

        return stage_candidates

    @transaction.atomic
    def get_remaining_not_final(self) -> int:
        return self._get_submissions_that_do_not_have_final_feedback().count()

    @transaction.atomic
    def get_available_in_stage(self) -> int:
        try:
            return self._get_available_submissions_in_subscription_stage().count()  # noqa
        except (SubscriptionTemporarilyEnded, SubscriptionEnded):
            return 0

    @transaction.atomic
    def get_or_create_work_assignment(self):
        taskqueryset = self._get_available_submissions_in_subscription_stage()
        task = get_random_element_from_queryset(taskqueryset)
        if self.assignments.filter(is_done=False).count() >= 2:
            raise NotMoreThanTwoOpenAssignmentsAllowed(
                'Not more than 2 active assignments allowed.')

        log.info(f'{self.owner} is assigned to {task} ({self.feedback_stage})')
        return TutorSubmissionAssignment.objects.create(
            subscription=self,
            submission=task.submission)

    @transaction.atomic
    def reserve_all_assignments_for_a_student(self):
        assert self.query_type == self.STUDENT_QUERY

        meta_submissions = self._get_submissions_that_do_not_have_final_feedback()  # noqa

        for meta in meta_submissions:
            submission = meta.submission
            if hasattr(submission, 'assignments'):
                submission.assignments.filter(is_done=False).delete()
            TutorSubmissionAssignment.objects.create(
                subscription=self,
                submission=submission
            )

        log.info(f'Loaded all subscriptions of student {self.query_key}')

    @transaction.atomic
    def delete(self):
        self.assignments.filter(is_done=False).delete()
        if self.assignments.count() == 0:
            super().delete()
        else:
            self.deactivated = True
            self.save()
